/**
 * @license
 * ReduxedChromeStorage v2.7.0
 * https://github.com/hindmost/reduxed-chrome-storage
 * Copyright (c) Savr Goryaev aka hindmost
 *
 * This source code is licensed under the MIT license
 * https://github.com/hindmost/reduxed-chrome-storage/blob/master/LICENSE
 */

function cloneDeep(o) {
    return o == null || typeof o !== 'object' ?
        o :
        JSON.parse(JSON.stringify(o));
}
function isEqual(a, b) {
    if (a === b)
        return true;
    if (a == null || typeof a !== 'object' ||
        b == null || typeof b !== 'object' ||
        Array.isArray(a) !== Array.isArray(b))
        return false;
    const keysA = Object.keys(a), keysB = Object.keys(b);
    if (keysA.length !== keysB.length)
        return false;
    for (const key of keysA) {
        if (keysB.indexOf(key) <= -1 || !isEqual(a[key], b[key]))
            return false;
    }
    return true;
}
function mergeOrReplace(a, b) {
    if (Array.isArray(b))
        return cloneDeep(b);
    if (a == null || typeof a !== 'object' || Array.isArray(a) ||
        b == null || typeof b !== 'object')
        return typeof b !== 'undefined' ? b : a;
    return Object.keys(a).concat(Object.keys(b).filter(key => !(key in a))).reduce((acc, key) => {
        return acc[key] = mergeOrReplace(a[key], b[key]), acc;
    }, {});
}

class ReduxedStorage {
    constructor({ createStore, reducer, storage, bufferLife, initialState, enhancer }) {
        this.createStore = createStore;
        this.storage = storage;
        this.reducer = reducer;
        this.enhancer = enhancer;
        this.buffLife = bufferLife ? Math.min(Math.max(bufferLife, 0), 2000) : 100;
        this.state0 = initialState;
        this.state = null;
        this.buffStore = null;
        this.lastState = null;
        this.listeners = [];
        this.inited = false;
        this.getState = this.getState.bind(this);
        this.dispatch = this.dispatch.bind(this);
        this.subscribe = this.subscribe.bind(this);
        this[Symbol.observable] = this[Symbol.observable].bind(this);
        this.replaceReducer = this.replaceReducer.bind(this);
    }
    init() {
        if (this.inited)
            return new Promise(resolve => {
                resolve(this);
            });
        const defaultState = this._createStore().getState();
        // subscribe for changes in chrome.storage
        this.storage.subscribe((data, oldData) => {
            if (isEqual(data, this.state))
                return;
            this._setState(data);
            for (const fn of this.listeners) {
                fn(oldData);
            }
        });
        this.inited = true;
        // return a promise to be resolved when the last state (if any) is restored from chrome.storage
        return new Promise(resolve => {
            // try to restore the last state stored in chrome.storage, if any
            this.storage.load(storedState => {
                let state = storedState ?
                    mergeOrReplace(defaultState, storedState) : defaultState;
                if (this.state0) {
                    state = mergeOrReplace(state, this.state0);
                }
                this._setState(state);
                if (!isEqual(state, storedState)) {
                    this._send2Storage(state);
                }
                resolve(this);
            });
        });
    }
    initFrom(state) {
        this._setState(state);
        this.inited = true;
        return this;
    }
    _createStore(initialState) {
        return this.createStore(this.reducer, initialState, this.enhancer);
    }
    _send2Storage(data) {
        this.storage.save(data);
    }
    _setState(data) {
        if (data) {
            this.state = cloneDeep(data);
        }
    }
    getState() {
        return this.state;
    }
    subscribe(fn) {
        typeof fn === 'function' && this.listeners.push(fn);
        return () => {
            if (typeof fn === 'function') {
                this.listeners = this.listeners.filter(v => v !== fn);
            }
        };
    }
    dispatch(action) {
        if (!this.buffStore) {
            // this.buffStore is to be used with sync actions
            this.buffStore = this._createStore(this.state);
            // this.lastState is shared by both sync and async actions
            this.lastState = this.buffStore.getState();
            setTimeout(() => {
                this.buffStore = null;
            }, this.buffLife);
        }
        // lastStore, holding an extra reference to the last created store, is to be used with async actions (e.g. via Redux Thunk);
        // then when this.buffStore is reset to null this variable should still refer to the same store
        let lastStore = this.buffStore;
        // set up a one-time state change listener
        const unsubscribe = lastStore.subscribe(() => {
            // if this.buffStore is non-empty, use it for getting the current state,
            // otherwise an async action is implied, so use lastStore instead
            const store = this.buffStore || lastStore;
            const state = store && store.getState();
            // we need a state change to be effective, so the current state should differ from the last saved one
            if (isEqual(state, this.lastState))
                return;
            // send the current state to chrome.storage & update this.lastState
            this._send2Storage(state);
            this.lastState = state;
            // unsubscribe this listener and reset the lastStore in order to release the related resources
            setTimeout(() => {
                unsubscribe();
                lastStore = null;
            }, this.buffLife);
        });
        return lastStore.dispatch(action);
    }
    replaceReducer(nextReducer) {
        if (typeof nextReducer === 'function') {
            this.reducer = nextReducer;
        }
        return this;
    }
    [Symbol.observable]() {
        const getState = this.getState;
        const subscribe = this.subscribe;
        return {
            subscribe(observer) {
                if (typeof observer !== 'object' || observer === null) {
                    throw new TypeError('Expected the observer to be an object.');
                }
                function observeState() {
                    const observerAsObserver = observer;
                    observerAsObserver.next && observerAsObserver.next(getState());
                }
                observeState();
                const unsubscribe = subscribe(observeState);
                return { unsubscribe };
            },
            [Symbol.observable]() {
                return this;
            }
        };
    }
}

var StorageAreaName;
(function (StorageAreaName) {
    StorageAreaName["local"] = "local";
    StorageAreaName["sync"] = "sync";
})(StorageAreaName || (StorageAreaName = {}));

const usageSize = (data) => new TextEncoder().encode(Object.entries(data).map(([key, val]) => key + JSON.stringify(val)).join('')).length;
class WrappedStorage {
    constructor({ namespace, area, key }) {
        this.ns = namespace;
        this.areaName = area === StorageAreaName.sync ? StorageAreaName.sync :
            StorageAreaName.local;
        this.key = key || 'reduxed';
        this.listeners = [];
        this.errorListeners = [];
    }
    init() {
        // Setup internal (shared) listener for chrome.storage.onChanged
        this.ns.storage.onChanged.addListener((changes, area) => {
            if (area !== this.areaName || !(this.key in changes))
                return;
            const { newValue, oldValue } = changes[this.key];
            if (!newValue)
                return;
            // call external chrome.storage.onChanged listeners
            for (const fn of this.listeners) {
                fn(newValue, oldValue);
            }
        });
    }
    subscribe(fn) {
        typeof fn === 'function' && this.listeners.push(fn);
    }
    subscribeForError(fn) {
        typeof fn === 'function' && this.errorListeners.push(fn);
    }
    fireErrorListeners(message, exceeded) {
        for (const fn of this.errorListeners) {
            fn(message, exceeded);
        }
    }
    callbackOnLoad(data, callback, all) {
        callback(!this.ns.runtime.lastError && (all ? data : data && data[this.key]));
    }
    callbackOnSave(data, area) {
        if (!this.ns.runtime.lastError)
            return;
        const { message } = this.ns.runtime.lastError;
        if (!message || !data || !area) {
            this.fireErrorListeners(message || '', false);
            return;
        }
        const b = this.areaName === StorageAreaName.sync &&
            area.QUOTA_BYTES_PER_ITEM &&
            usageSize({ [this.key]: data }) > area.QUOTA_BYTES_PER_ITEM;
        if (b) {
            this.fireErrorListeners(message, true);
            return;
        }
        this.load((allData) => {
            const b = typeof allData === 'object' &&
                area.QUOTA_BYTES > 0 &&
                usageSize(Object.assign(Object.assign({}, allData), { [this.key]: data })) > area.QUOTA_BYTES;
            this.fireErrorListeners(message, b);
        }, true);
    }
}

class WrappedChromeStorage extends WrappedStorage {
    constructor({ namespace, area, key }) {
        super({ namespace, area, key });
        this.areaApi = this.ns.storage[this.areaName];
    }
    load(fn, all) {
        typeof fn === 'function' &&
            this.areaApi.get(all ? null : this.key, data => {
                this.callbackOnLoad(data, fn, all);
            });
    }
    save(data) {
        this.areaApi.set({ [this.key]: data }, () => {
            this.callbackOnSave(data, this.areaApi);
        });
    }
}

class WrappedBrowserStorage extends WrappedStorage {
    constructor({ namespace, area, key }) {
        super({ namespace, area, key });
        this.areaApi = this.ns.storage[this.areaName];
    }
    load(fn, all) {
        typeof fn === 'function' &&
            this.areaApi.get(all ? null : this.key).then(data => {
                this.callbackOnLoad(data, fn);
            });
    }
    save(data) {
        this.areaApi.set({ [this.key]: data }).then(() => {
            this.callbackOnSave(data, this.areaApi);
        });
    }
}

var Namespace;
(function (Namespace) {
    Namespace["chrome"] = "chrome";
    Namespace["browser"] = "browser";
})(Namespace || (Namespace = {}));
/**
 * ReduxedChromeStorage creator factory.
 * Returns an async store creator that's supposed to replace
 * the original Redux's createStore function.
 * Unlike the original createStore() that immediately returns a store,
 * async store creator returns a Promise to be resolved
 * when the created store is ready
 * @param obj
 * @param obj.createStore the original Redux's createStore function.
 * The only mandatory property/option
 * @param obj.namespace string to identify the APIs namespace to be used,
 * either 'chrome' or 'browser'.
 * If this and the next two options are missing,
 * the chrome namespace is used by default
 * @param obj.chromeNs the chrome namespace within Manifest V2 extension.
 * If this option is supplied, the previous one is ignored
 * @param obj.browserNs the browser namespace within Firefox extension,
 * or the chrome namespace within Manifest V3 chrome extension.
 * If this option is supplied, the previous two are ignored
 * @param obj.changeListener a function to be called whenever the state changes,
 * receives two parameters:
 * 1) a one-time store - container of the current state;
 * 2) the previous state.
 * If this option is supplied, the async store creator returned by the factory
 * is not supposed to be immediately used for store creation
 * @param obj.errorListener a function to be called whenever an error occurs
 * during chrome.storage update, receives two parameters:
 * 1) an error message defined by storage API;
 * 2) a boolean indicating if the limit for the used storage area is exceeded
 * @param obj.storageArea the name of chrome.storage area to be used,
 * either 'local' or 'sync', defaults to 'local'
 * @param obj.storageKey key under which the state will be stored/tracked
 * in chrome.storage, defaults to 'reduxed'
 * @param obj.bufferLife lifetime of the bulk actions buffer (in ms),
 * defaults to 100
 * @returns an async store creator to replace the original createStore function
 */
function reduxedStorageCreatorFactory({ createStore, namespace, chromeNs, browserNs, changeListener, errorListener, storageArea, storageKey, bufferLife }) {
    if (typeof createStore !== 'function')
        throw new Error(`Missing 'createStore' property/option`);
    const storage = browserNs || namespace === Namespace.browser ?
        new WrappedBrowserStorage({
            namespace: browserNs || browser, area: storageArea, key: storageKey
        }) :
        new WrappedChromeStorage({
            namespace: chromeNs || chrome, area: storageArea, key: storageKey
        });
    storage.init();
    typeof errorListener === 'function' &&
        storage.subscribeForError(errorListener);
    function asyncStoreCreator(reducer, initialState, enhancer) {
        if (typeof reducer !== 'function')
            throw new Error(`Missing 'reducer' parameter`);
        if (typeof initialState === 'function' && typeof enhancer === 'function')
            throw new Error(`Multiple 'enhancer' parameters unallowed`);
        if (typeof initialState === 'function' && typeof enhancer === 'undefined') {
            enhancer = initialState;
            initialState = undefined;
        }
        const opts = {
            createStore, storage, bufferLife, reducer, initialState, enhancer
        };
        if (typeof changeListener === 'function') {
            storage.subscribe((data, oldData) => {
                const store = new ReduxedStorage(opts);
                changeListener(store.initFrom(data), oldData);
            });
            return new Promise(resolve => {
                resolve(createStore(state => state));
            });
        }
        const store = new ReduxedStorage(opts);
        return store.init();
    }
    return asyncStoreCreator;
}

export { reduxedStorageCreatorFactory as default };
